"""
 Unit tests for the Roman jump step code
"""

from itertools import cycle

import numpy as np
import pytest
import roman_datamodels.stnode as rds
from astropy import units as u
from astropy.time import Time
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils
from roman_datamodels.datamodels import GainRefModel, ReadnoiseRefModel

from romancal.jump import JumpStep

RNG = np.random.default_rng(42)

MAXIMUM_CORES = ["none", "quarter", "half", "all"]


@pytest.fixture(scope="module")
def generate_wfi_reffiles(tmp_path_factory):
    gainfile = str(tmp_path_factory.mktemp("ndata") / "gain.asdf")
    readnoisefile = str(tmp_path_factory.mktemp("ndata") / "readnoise.asdf")

    ingain = 6
    xsize = 20
    ysize = 20

    shape = (ysize, xsize)

    # Create temporary gain reference file
    gain_ref = rds.GainRef()
    meta = maker_utils.mk_ref_common("GAIN")
    meta["instrument"]["detector"] = "WFI01"
    meta["instrument"]["name"] = "WFI"
    meta["author"] = "John Doe"
    meta["pedigree"] = "DUMMY"
    meta["description"] = "DUMMY"
    meta["useafter"] = Time("2022-01-01T11:11:11.111")

    gain_ref["meta"] = meta
    gain_ref["data"] = u.Quantity(
        np.ones(shape, dtype=np.float32) * ingain, u.electron / u.DN, dtype=np.float32
    )
    gain_ref["dq"] = np.zeros(shape, dtype=np.uint16)
    gain_ref["err"] = u.Quantity(
        (RNG.uniform(size=shape) * 0.05).astype(np.float64),
        u.electron / u.DN,
        dtype=np.float64,
    )

    gain_ref_model = GainRefModel(gain_ref)
    gain_ref_model.save(gainfile)

    # Create temporary readnoise reference file
    rn_ref = rds.ReadnoiseRef()
    meta = maker_utils.mk_ref_common("READNOISE")
    meta["instrument"]["detector"] = "WFI01"
    meta["instrument"]["name"] = "WFI"
    meta["author"] = "John Doe"
    meta["pedigree"] = "DUMMY"
    meta["description"] = "DUMMY"
    meta["useafter"] = Time("2022-01-01T11:11:11.111")
    meta["exposure"] = {}
    meta["exposure"]["type"] = "WFI_IMAGE"
    meta["exposure"]["p_exptype"] = "WFI_IMAGE|WFI_GRISM|WFI_PRISM|"

    rn_ref["meta"] = meta
    rn_ref["data"] = u.Quantity(
        np.ones(shape, dtype=np.float32), u.DN, dtype=np.float32
    )
    rn_ref["dq"] = np.zeros(shape, dtype=np.uint16)
    rn_ref["err"] = u.Quantity(
        (RNG.uniform(size=shape) * 0.05).astype(np.float64), u.DN, dtype=np.float64
    )

    rn_ref_model = ReadnoiseRefModel(rn_ref)
    rn_ref_model.save(readnoisefile)

    return gainfile, readnoisefile


@pytest.fixture
def setup_inputs():
    def _setup(
        ngroups=10,
        readnoise=10,
        nrows=20,
        ncols=20,
        nframes=1,
        grouptime=1.0,
        gain=1,
        deltatime=1,
    ):
        err = np.ones(shape=(ngroups, nrows, ncols), dtype=np.float32)
        data = np.zeros(shape=(ngroups, nrows, ncols), dtype=np.float32)
        gdq = np.zeros(shape=(ngroups, nrows, ncols), dtype=np.uint8)
        pixdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)

        csize = (ngroups, nrows, ncols)
        dm_ramp = rdm.RampModel(maker_utils.mk_ramp(shape=csize))

        dm_ramp.meta.instrument.name = "WFI"
        dm_ramp.meta.instrument.optical_element = "F158"

        dm_ramp.data = u.Quantity(data + 6.0, u.DN, dtype=np.float32)
        dm_ramp.pixeldq = pixdq
        dm_ramp.groupdq = gdq
        dm_ramp.err = u.Quantity(err, u.DN, dtype=np.float32)

        dm_ramp.meta.exposure.type = "WFI_IMAGE"
        dm_ramp.meta.exposure.group_time = deltatime
        dm_ramp.meta.exposure.frame_time = deltatime
        dm_ramp.meta.exposure.ngroups = ngroups
        dm_ramp.meta.exposure.nframes = 1
        dm_ramp.meta.exposure.groupgap = 0
        dm_ramp.meta.cal_step["dq_init"] = "INCOMPLETE"
        dm_ramp.meta.cal_step["jump"] = "INCOMPLETE"

        return dm_ramp

    return _setup


@pytest.mark.parametrize("max_cores", MAXIMUM_CORES)
def test_one_CR(generate_wfi_reffiles, max_cores, setup_inputs):
    override_gain, override_readnoise = generate_wfi_reffiles

    grouptime = 3.0
    deltaDN = 5
    ingain = 200
    inreadnoise = 7.0
    ngroups = 100
    CR_fraction = 3
    xsize = 20
    ysize = 20

    model1 = setup_inputs(
        ngroups=ngroups,
        nrows=ysize,
        ncols=xsize,
        gain=ingain,
        readnoise=inreadnoise,
        deltatime=grouptime,
    )

    for i in range(ngroups):
        model1.data[i, :, :] = deltaDN * i * model1.data.unit

    first_CR_group_locs = [x for x in range(1, 7) if x % 5 == 0]

    CR_locs = [x for x in range(xsize * ysize) if x % CR_fraction == 0]
    CR_x_locs = [x % ysize for x in CR_locs]
    CR_y_locs = [int(x / xsize) for x in CR_locs]
    CR_pool = cycle(first_CR_group_locs)

    # Add CRs to the SCI data
    for i in range(len(CR_x_locs)):
        CR_group = next(CR_pool)
        model1.data[CR_group:, CR_y_locs[i], CR_x_locs[i]] = (
            model1.data[CR_group:, CR_y_locs[i], CR_x_locs[i]]
            + 500.0 * model1.data.unit
        )

    out_model = JumpStep.call(
        model1,
        override_gain=override_gain,
        override_readnoise=override_readnoise,
        maximum_cores=max_cores,
        use_ramp_jump_detection=False,
    )

    CR_pool = cycle(first_CR_group_locs)
    for i in range(len(CR_x_locs)):
        CR_group = next(CR_pool)

        assert 4 == np.max(out_model.groupdq[CR_group, CR_y_locs[i], CR_x_locs[i]])


@pytest.mark.parametrize("max_cores", MAXIMUM_CORES)
def test_two_CRs(generate_wfi_reffiles, max_cores, setup_inputs):
    override_gain, override_readnoise = generate_wfi_reffiles

    grouptime = 3.0
    deltaDN = 5
    ingain = 6
    inreadnoise = 7.0
    ngroups = 100
    CR_fraction = 5
    xsize = 20
    ysize = 20

    model1 = setup_inputs(
        ngroups=ngroups,
        nrows=ysize,
        ncols=xsize,
        gain=ingain,
        readnoise=inreadnoise,
        deltatime=grouptime,
    )

    for i in range(ngroups):
        model1.data[i, :, :] = deltaDN * i * model1.data.unit

    first_CR_group_locs = [x for x in range(1, 7) if x % 5 == 0]
    CR_locs = [x for x in range(xsize * ysize) if x % CR_fraction == 0]
    CR_x_locs = [x % ysize for x in CR_locs]
    CR_y_locs = [int(x / xsize) for x in CR_locs]
    CR_pool = cycle(first_CR_group_locs)

    for i in range(len(CR_x_locs)):
        CR_group = next(CR_pool)

        model1.data[CR_group:, CR_y_locs[i], CR_x_locs[i]] = (
            model1.data[CR_group:, CR_y_locs[i], CR_x_locs[i]] + 500 * model1.data.unit
        )
        model1.data[CR_group + 8 :, CR_y_locs[i], CR_x_locs[i]] = (
            model1.data[CR_group + 8 :, CR_y_locs[i], CR_x_locs[i]]
            + 700 * model1.data.unit
        )

    out_model = JumpStep.call(
        model1,
        override_gain=override_gain,
        override_readnoise=override_readnoise,
        maximum_cores=max_cores,
        use_ramp_jump_detection=False,
    )

    CR_pool = cycle(first_CR_group_locs)
    for i in range(len(CR_x_locs)):
        CR_group = next(CR_pool)

        assert 4 == np.max(out_model.groupdq[CR_group, CR_y_locs[i], CR_x_locs[i]])
        assert 4 == np.max(out_model.groupdq[CR_group + 8, CR_y_locs[i], CR_x_locs[i]])


@pytest.mark.parametrize("max_cores", MAXIMUM_CORES)
def test_two_group_integration(generate_wfi_reffiles, max_cores, setup_inputs):
    override_gain, override_readnoise = generate_wfi_reffiles
    grouptime = 3.0
    ingain = 6
    inreadnoise = 7
    ngroups = 2
    xsize = 20
    ysize = 20
    model1 = setup_inputs(
        ngroups=ngroups,
        nrows=ysize,
        ncols=xsize,
        gain=ingain,
        readnoise=inreadnoise,
        deltatime=grouptime,
    )

    out_model = JumpStep.call(
        model1,
        override_gain=override_gain,
        override_readnoise=override_readnoise,
        maximum_cores=max_cores,
    )

    assert out_model.meta.cal_step["jump"] == "SKIPPED"


@pytest.mark.parametrize("use_jump", [False, True])
def test_four_group_integration(generate_wfi_reffiles, setup_inputs, use_jump):
    override_gain, override_readnoise = generate_wfi_reffiles
    grouptime = 3.0
    ingain = 6
    inreadnoise = 7
    ngroups = 4
    xsize = 20
    ysize = 20
    model1 = setup_inputs(
        ngroups=ngroups,
        nrows=ysize,
        ncols=xsize,
        gain=ingain,
        readnoise=inreadnoise,
        deltatime=grouptime,
    )

    out_model = JumpStep.call(
        model1,
        override_gain=override_gain,
        override_readnoise=override_readnoise,
        maximum_cores="none",
        use_ramp_jump_detection=use_jump,
    )

    if use_jump:
        assert out_model.meta.cal_step["jump"] == "SKIPPED"
    else:
        assert out_model.meta.cal_step["jump"] == "COMPLETE"


@pytest.mark.parametrize("use_jump", [False, True])
def test_three_group_integration(generate_wfi_reffiles, setup_inputs, use_jump):
    override_gain, override_readnoise = generate_wfi_reffiles
    grouptime = 3.0
    ingain = 6
    inreadnoise = 7
    ngroups = 3
    xsize = 20
    ysize = 20
    model1 = setup_inputs(
        ngroups=ngroups,
        nrows=ysize,
        ncols=xsize,
        gain=ingain,
        readnoise=inreadnoise,
        deltatime=grouptime,
    )

    out_model = JumpStep.call(
        model1,
        override_gain=override_gain,
        override_readnoise=override_readnoise,
        maximum_cores="none",
        use_ramp_jump_detection=use_jump,
    )

    if use_jump:
        assert out_model.meta.cal_step.jump == "SKIPPED"
    else:
        assert out_model.meta.cal_step.jump == "COMPLETE"



"""Association Generator

The Association Generator takes a list of items, an Association Pool, and
creates sub-lists of those items depending on each item's attributes. How the
sub-lists are created is defined by Association Rules.

For more, see the :ref:`documentation overview <asn-overview>`.

"""

# flake8: noqa: F401, F402, F403

# Take version from the upstream package
from .. import __version__


# Utility
def libpath(filepath):
    """Return the full path to the module library."""
    from os.path import abspath, dirname, join

    return join(dirname(abspath(__file__)), "lib", filepath)


from .association import *
from .association_io import *
from .exceptions import *
from .generate import *
from .lib.process_list import *
from .load_asn import load_asn
from .main import *
from .pool import *
from .registry import *



"""Association Member"""

from collections import UserDict
from copy import copy


class Member(UserDict):
    """Member of an association

    Parameters
    ----------
    initialdata: Dict-like or Member
        Initialization data. Any type of initialization that
        `collections.UserDict` allows or `Member` itself.

    item: obj
        The item to initialize with. This will override
        any `Member.item` given in `initialdata`.

    Attributes
    ----------
    item: obj
        The original item that created this member.
    """

    def __init__(self, initialdata=None, item=None):
        self.item = None

        if isinstance(initialdata, Member):
            self.data = copy(initialdata.data)
            self.item = copy(initialdata.item)
        else:
            super().__init__(initialdata)

        if item is not None:
            self.item = copy(item)



"""
Flat-field a science image
"""

import roman_datamodels as rdm

from ..stpipe import RomanStep
from . import flat_field

__all__ = ["FlatFieldStep"]


class FlatFieldStep(RomanStep):
    """Flat-field a science image using a flatfield reference image."""

    reference_file_types = ["flat"]

    def process(self, input_model):
        input_model = rdm.open(input_model, lazy_load=False)

        reference_file_name = self.get_reference_file(input_model, "flat")

        # Check for a valid reference file
        if reference_file_name == "N/A":
            self.log.warning("No FLAT reference file found")
            self.log.warning("Flat Field step will be skipped")
            reference_file_name = None

        if reference_file_name is not None:
            reference_file_model = rdm.open(reference_file_name)
            self.log.debug(f"Using FLAT ref file: {reference_file_name}")
        else:
            reference_file_model = None
            self.log.debug("Using no FLAT ref file")

        # Do the flat-field correction
        output_model = flat_field.do_correction(
            input_model,
            reference_file_model,
        )

        # Close the input and reference files
        input_model.close()
        try:
            reference_file_model.close()
        except AttributeError:
            pass

        if self.save_results:
            try:
                self.suffix = "flat"
            except AttributeError:
                self["suffix"] = "flat"

        return output_model



import logging
from typing import List

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from drizzle import cdrizzle, util
from roman_datamodels import datamodels, maker_utils, stnode
from stcal.alignment.util import compute_scale

from ..assign_wcs import utils
from ..datamodels import ModelContainer
from . import gwcs_drizzle, resample_utils

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["OutputTooLargeError", "ResampleData"]


class OutputTooLargeError(MemoryError):
    """Raised when the output is too large for in-memory instantiation"""


class ResampleData:
    """
    This is the controlling routine for the resampling process.

    Notes
    -----
    This routine performs the following operations:

      1. Extracts parameter settings from input model, such as pixfrac,
         weight type, exposure time (if relevant), and kernel, and merges
         them with any user-provided values.
      2. Creates output WCS based on input images and define mapping function
         between all input arrays and the output array. Alternatively, a custom,
         user-provided WCS object can be used instead.
      3. Updates output data model with output arrays from drizzle, including
         a record of metadata from all input models.
    """

    def __init__(
        self,
        input_models,
        output=None,
        single=False,
        blendheaders=True,
        pixfrac=1.0,
        kernel="square",
        fillval="INDEF",
        wht_type="ivm",
        good_bits="0",
        pscale_ratio=1.0,
        pscale=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        input_models : list of objects
            list of data models, one for each input image

        output : str
            filename for output

        kwargs : dict
            Other parameters.

            .. note::
                ``output_shape`` is in the ``x, y`` order.

            .. note::
                ``in_memory`` controls whether or not the resampled
                array from ``resample_many_to_many()``
                should be kept in memory or written out to disk and
                deleted from memory. Default value is `True` to keep
                all products in memory.
        """
        if (
            (input_models is None)
            or (len(input_models) == 0)
            or (not any(input_models))
        ):
            raise ValueError(
                "No input has been provided. Input should be a list of datamodel(s) or "
                "a ModelContainer."
            )

        self.input_models = input_models
        self.output_filename = output
        self.pscale_ratio = pscale_ratio
        self.single = single
        self.blendheaders = blendheaders
        self.pixfrac = pixfrac
        self.kernel = kernel
        self.fillval = fillval
        self.weight_type = wht_type
        self.good_bits = good_bits
        self.in_memory = kwargs.get("in_memory", True)

        log.info(f"Driz parameter kernel: {self.kernel}")
        log.info(f"Driz parameter pixfrac: {self.pixfrac}")
        log.info(f"Driz parameter fillval: {self.fillval}")
        log.info(f"Driz parameter weight_type: {self.weight_type}")

        output_wcs = kwargs.get("output_wcs", None)
        output_shape = kwargs.get("output_shape", None)
        crpix = kwargs.get("crpix", None)
        crval = kwargs.get("crval", None)
        rotation = kwargs.get("rotation", None)

        if pscale is not None:
            log.info(f"Output pixel scale: {pscale} arcsec.")
            pscale /= 3600.0
        else:
            log.info(f"Output pixel scale ratio: {pscale_ratio}")

        # build the output WCS object
        if output_wcs:
            # use the provided WCS object
            self.output_wcs = output_wcs
            if output_shape is not None:
                self.output_wcs.array_shape = output_shape[::-1]
        else:
            # determine output WCS based on all inputs, including a reference WCS
            self.output_wcs = resample_utils.make_output_wcs(
                self.input_models,
                pscale_ratio=self.pscale_ratio,
                pscale=pscale,
                rotation=rotation,
                shape=None if output_shape is None else output_shape[::-1],
                crpix=crpix,
                crval=crval,
            )

        log.debug(f"Output mosaic size: {self.output_wcs.array_shape}")

        # NOTE: should we enable memory allocation?

        # can_allocate, required_memory = datamodels.util.check_memory_allocation(
        #     self.output_wcs.array_shape,
        #     kwargs['allowed_memory'],
        #     datamodels.ImageModel
        # )
        # if not can_allocate:
        #     raise OutputTooLargeError(
        #         f'Combined ImageModel size {self.output_wcs.array_shape} '
        #         f'requires {bytes2human(required_memory)}. '
        #         f'Model cannot be instantiated.'
        #     )

        # NOTE: wait for William to fix bug in datamodels' init and then
        # use datamodels.ImageModel(shape=(nx, ny)) instead of mk_datamodel()

        self.blank_output = maker_utils.mk_datamodel(
            datamodels.MosaicModel, shape=tuple(self.output_wcs.array_shape)
        )

        # update meta.basic
        populate_mosaic_basic(self.blank_output, input_models)

        # update meta.cal_step
        self.blank_output.meta.cal_step = maker_utils.mk_l3_cal_step(
            **input_models[0].meta.cal_step.to_flat_dict()
        )

        # Update the output with all the component metas
        populate_mosaic_individual(self.blank_output, input_models)

        # update meta data and wcs
        # note we have made this input_model_0 variable so that if
        # meta includes lazily-loaded objects, that we can successfully
        # copy them into the metadata.  Directly running input_models[0].meta
        # below can lead to input_models[0] going out of scope after
        # meta is loaded but before the dictionary is constructed,
        # which can lead to seek on closed file errors if
        # meta contains lazily loaded objects.
        input_model_0 = input_models[0]
        l2_into_l3_meta(self.blank_output.meta, input_model_0.meta)
        self.blank_output.meta.wcs = self.output_wcs
        gwcs_into_l3(self.blank_output, self.output_wcs)
        self.blank_output.cal_logs = stnode.CalLogs()
        self.blank_output["individual_image_cal_logs"] = [
            model.cal_logs for model in input_models
        ]

        self.output_models = ModelContainer()

    def do_drizzle(self):
        """Pick the correct drizzling mode based on ``self.single``."""
        if self.single:
            return self.resample_many_to_many()
        else:
            return self.resample_many_to_one()

    def resample_many_to_many(self):
        """Resample many inputs to many outputs where outputs have a common frame.

        Coadd only different detectors of the same exposure (e.g. map SCA 1 and
        10 onto the same output image), as they image different areas of the
        sky.

        Used for outlier detection
        """
        for exposure in self.input_models.models_grouped:
            output_model = self.blank_output
            output_model.meta["resample"] = maker_utils.mk_resample()
            # Determine output file type from input exposure filenames
            # Use this for defining the output filename
            indx = exposure[0].meta.filename.rfind(".")
            output_type = exposure[0].meta.filename[indx:]
            output_root = "_".join(
                exposure[0].meta.filename.replace(output_type, "").split("_")[:-1]
            )
            output_model.meta.filename = f"{output_root}_outlier_i2d{output_type}"

            # Initialize the output with the wcs
            driz = gwcs_drizzle.GWCSDrizzle(
                output_model,
                pixfrac=self.pixfrac,
                kernel=self.kernel,
                fillval=self.fillval,
            )

            log.info(f"{len(exposure)} exposures to drizzle together")
            output_list = []
            for img in exposure:
                img = datamodels.open(img)
                # TODO: should weight_type=None here?
                inwht = resample_utils.build_driz_weight(
                    img, weight_type=self.weight_type, good_bits=self.good_bits
                )

                # apply sky subtraction
                if not hasattr(img.meta, "background"):
                    self._create_background_attribute(img)
                blevel = img.meta.background.level
                if not img.meta.background.subtracted and blevel is not None:
                    data = img.data - blevel
                else:
                    data = img.data

                xmin, xmax, ymin, ymax = resample_utils.resample_range(
                    data.shape, img.meta.wcs.bounding_box
                )

                driz.add_image(
                    data,
                    img.meta.wcs,
                    inwht=inwht,
                    xmin=xmin,
                    xmax=xmax,
                    ymin=ymin,
                    ymax=ymax,
                )
                del data
                img.close()

            # cast context array to uint32
            output_model.context = output_model.context.astype("uint32")
            if not self.in_memory:
                # Write out model to disk, then return filename
                output_name = output_model.meta.filename
                output_model.save(output_name)
                log.info(f"Exposure {output_name} saved to file")
                output_list.append(output_name)
            else:
                output_list.append(output_model.copy())

            self.output_models = ModelContainer(output_list, return_open=self.in_memory)
            output_model.data *= 0.0
            output_model.weight *= 0.0

        return self.output_models

    def resample_many_to_one(self):
        """Resample and coadd many inputs to a single output.
        Used for level 3 resampling
        """
        output_model = self.blank_output.copy()
        output_model.meta.filename = self.output_filename
        output_model.meta["resample"] = maker_utils.mk_resample()
        output_model.meta.resample["members"] = []
        output_model.meta.resample.weight_type = self.weight_type
        output_model.meta.resample.pointings = len(self.input_models.models_grouped)

        if self.blendheaders:
            log.info("Skipping blendheaders for now.")

        # Initialize the output with the wcs
        driz = gwcs_drizzle.GWCSDrizzle(
            output_model,
            outwcs=self.output_wcs,
            pixfrac=self.pixfrac,
            kernel=self.kernel,
            fillval=self.fillval,
        )

        log.info("Resampling science data")
        members = []
        for img in self.input_models:
            inwht = resample_utils.build_driz_weight(
                img,
                weight_type=self.weight_type,
                good_bits=self.good_bits,
            )
            if not hasattr(img.meta, "background"):
                self._create_background_attribute(img)
            blevel = img.meta.background.level
            if not img.meta.background.subtracted and blevel is not None:
                data = img.data - blevel
            else:
                data = img.data

            xmin, xmax, ymin, ymax = resample_utils.resample_range(
                data.shape, img.meta.wcs.bounding_box
            )

            driz.add_image(
                data,
                img.meta.wcs,
                inwht=inwht,
                xmin=xmin,
                xmax=xmax,
                ymin=ymin,
                ymax=ymax,
            )
            del data, inwht
            members.append(str(img.meta.filename))

        members = (
            members
            if self.input_models.filepaths is None
            else self.input_models.filepaths
        )
        output_model.meta.resample.members = members

        # Resample variances array in self.input_models to output_model
        self.resample_variance_array("var_rnoise", output_model)
        self.resample_variance_array("var_poisson", output_model)
        self.resample_variance_array("var_flat", output_model)

        # Make exposure time image
        exptime_tot = self.resample_exposure_time(output_model)

        # TODO: fix unit here
        output_model.err = u.Quantity(
            np.sqrt(
                np.nansum(
                    [
                        output_model.var_rnoise,
                        output_model.var_poisson,
                        output_model.var_flat,
                    ],
                    axis=0,
                )
            ),
            unit=output_model.err.unit,
        )

        self.update_exposure_times(output_model, exptime_tot)

        # TODO: fix RAD to expect a context image datatype of int32
        output_model.context = output_model.context.astype(np.uint32)

        self.output_models.append(output_model)

        return self.output_models

    def _create_background_attribute(self, img):
        img.meta["background"] = {}
        img.meta.background["level"] = 0
        img.meta.background["subtracted"] = True

    def resample_variance_array(self, name, output_model):
        """Resample variance arrays from ``self.input_models`` to the ``output_model``.

        Resample the ``name`` variance array to the same name in ``output_model``,
        using a cumulative sum.

        This modifies ``output_model`` in-place.
        """
        output_wcs = self.output_wcs
        inverse_variance_sum = np.full_like(output_model.data.value, np.nan)

        log.info(f"Resampling {name}")
        for model in self.input_models:
            variance = getattr(model, name)
            if variance is None or variance.size == 0:
                log.debug(
                    f"No data for '{name}' for model "
                    f"{repr(model.meta.filename)}. Skipping ..."
                )
                continue
            elif variance.shape != model.data.shape:
                log.warning(
                    f"Data shape mismatch for '{name}' for model "
                    f"{repr(model.meta.filename)}. Skipping..."
                )
                continue

            # create a unit weight map for all the input pixels with science data
            inwht = resample_utils.build_driz_weight(
                model, weight_type=None, good_bits=self.good_bits
            )

            resampled_variance = np.zeros_like(output_model.data)
            outwht = np.zeros_like(output_model.data)
            outcon = np.zeros_like(output_model.context)

            xmin, xmax, ymin, ymax = resample_utils.resample_range(
                variance.shape, model.meta.wcs.bounding_box
            )

            # resample the variance array (fill "unpopulated" pixels with NaNs)
            self.drizzle_arrays(
                variance,
                inwht,
                model.meta.wcs,
                output_wcs,
                resampled_variance,
                outwht,
                outcon,
                pixfrac=self.pixfrac,
                kernel=self.kernel,
                fillval=np.nan,
                xmin=xmin,
                xmax=xmax,
                ymin=ymin,
                ymax=ymax,
            )

            # Add the inverse of the resampled variance to a running sum.
            # Update only pixels (in the running sum) with valid new values:
            mask = resampled_variance > 0

            inverse_variance_sum[mask] = np.nansum(
                [
                    inverse_variance_sum[mask],
                    np.reciprocal(resampled_variance[mask]),
                ],
                axis=0,
            )

        # We now have a sum of the inverse resampled variances.  We need the
        # inverse of that to get back to units of variance.
        # TODO: fix unit here
        output_variance = u.Quantity(
            np.reciprocal(inverse_variance_sum), unit=u.MJy**2 / u.sr**2
        )

        setattr(output_model, name, output_variance)

    def resample_exposure_time(self, output_model):
        """Resample the exposure time from ``self.input_models`` to the
        ``output_model``.

        Create an exposure time image that is the drizzled sum of the input
        images.
        """
        output_wcs = self.output_wcs
        exptime_tot = np.zeros(output_model.data.shape, dtype="f4")

        log.info("Resampling exposure time")
        for model in self.input_models:
            exptime = np.full(
                model.data.shape, model.meta.exposure.effective_exposure_time
            )

            # create a unit weight map for all the input pixels with science data
            inwht = resample_utils.build_driz_weight(
                model, weight_type=None, good_bits=self.good_bits
            )

            resampled_exptime = np.zeros_like(output_model.data)
            outwht = np.zeros_like(output_model.data)
            outcon = np.zeros_like(output_model.context, dtype="i4")
            # drizzle wants an i4, but datamodels wants a u4.

            xmin, xmax, ymin, ymax = resample_utils.resample_range(
                exptime.shape, model.meta.wcs.bounding_box
            )

            # resample the exptime array
            self.drizzle_arrays(
                exptime * u.s,  # drizzle_arrays expects these to have units
                inwht,
                model.meta.wcs,
                output_wcs,
                resampled_exptime,
                outwht,
                outcon,
                pixfrac=1,  # for exposure time images, always use pixfrac = 1
                kernel=self.kernel,
                fillval=0,
                xmin=xmin,
                xmax=xmax,
                ymin=ymin,
                ymax=ymax,
            )

            exptime_tot += resampled_exptime.value

        return exptime_tot

    def update_exposure_times(self, output_model, exptime_tot):
        """Update exposure time metadata (in-place)."""
        m = exptime_tot > 0
        total_exposure_time = np.mean(exptime_tot[m]) if np.any(m) else 0
        max_exposure_time = np.max(exptime_tot)
        log.info(
            f"Mean, max exposure times: {total_exposure_time:.1f}, "
            f"{max_exposure_time:.1f}"
        )
        exposure_times = {"start": [], "end": []}
        for exposure in self.input_models.models_grouped:
            exposure_times["start"].append(exposure[0].meta.exposure.start_time)
            exposure_times["end"].append(exposure[0].meta.exposure.end_time)

        # Update some basic exposure time values based on output_model
        output_model.meta.basic.mean_exposure_time = total_exposure_time
        output_model.meta.basic.time_first_mjd = min(exposure_times["start"]).mjd
        output_model.meta.basic.time_last_mjd = max(exposure_times["end"]).mjd
        output_model.meta.basic.max_exposure_time = max_exposure_time
        output_model.meta.resample.product_exposure_time = max_exposure_time

    @staticmethod
    def drizzle_arrays(
        insci,
        inwht,
        input_wcs,
        output_wcs,
        outsci,
        outwht,
        outcon,
        uniqid=1,
        xmin=None,
        xmax=None,
        ymin=None,
        ymax=None,
        pixfrac=1.0,
        kernel="square",
        fillval="INDEF",
        wtscale=1.0,
    ):
        """
        Low level routine for performing 'drizzle' operation on one image.

        The interface is compatible with STScI code. All images are Python
        `ndarrays`, instead of filenames. File handling (input and output) is
        performed by the calling routine.

        Parameters
        ----------

        insci : 2d array
            A 2d `numpy` array containing the input image to be drizzled.

        inwht : 2d array
            A 2d `numpy` array containing the pixel by pixel weighting.
            Must have the same dimensions as insci. If none is supplied,
            the weighting is set to one.

        input_wcs : `gwcs.wcs.WCS` object
            The world coordinate system of the input image.

        output_wcs : `gwcs.wcs.WCS` object
            The world coordinate system of the output image.

        outsci : 2d array
            A 2d `numpy` array containing the output image produced by
            drizzling. On the first call it should be set to zero.
            Subsequent calls it will hold the intermediate results.  This
            is modified in-place.

        outwht : 2d array
            A 2d `numpy` array containing the output counts. On the first
            call it should be set to zero. On subsequent calls it will
            hold the intermediate results.  This is modified in-place.

        outcon : 2d or 3d array, optional
            A 2d or 3d `numpy` array holding a bitmap of which image was an input
            for each output pixel. Should be integer zero on first call.
            Subsequent calls hold intermediate results.  This is modified
            in-place.

        uniqid : int, optional
            The id number of the input image. Should be one the first time
            this function is called and incremented by one on each subsequent
            call.

        xmin : float, None, optional
            on the input image. Only pixels on the input image inside this
            rectangle will have their flux added to the output image. Xmin
            sets the minimum value of the x dimension. The x dimension is the
            dimension that varies quickest on the image. If the value is zero,
            no minimum will be set in the x dimension. All four parameters are
            zero based, counting starts at zero.

        xmax : float, None, optional
            Sets the maximum value of the x dimension on the bounding box
            of the input image. If the value is zero, no maximum will
            be set in the x dimension, the full x dimension of the output
            image is the bounding box.

        ymin : float, None, optional
            Sets the minimum value in the y dimension on the bounding box. The
            y dimension varies less rapidly than the x and represents the line
            index on the input image. If the value is zero, no minimum  will be
            set in the y dimension.

        ymax : float, None, optional
            Sets the maximum value in the y dimension. If the value is zero, no
            maximum will be set in the y dimension,  the full x dimension
            of the output image is the bounding box.

        pixfrac : float, optional
            The fraction of a pixel that the pixel flux is confined to. The
            default value of 1 has the pixel flux evenly spread across the image.
            A value of 0.5 confines it to half a pixel in the linear dimension,
            so the flux is confined to a quarter of the pixel area when the square
            kernel is used.

        kernel: str, optional
            The name of the kernel used to combine the input. The choice of
            kernel controls the distribution of flux over the kernel. The kernel
            names are: `'square'`, `'gaussian'`, `'point'`, `'tophat'`, `'turbo'`,
            `'lanczos2'`, and `'lanczos3'`. The `'square'` kernel is the default.

        fillval: str, optional
            The value a pixel is set to in the output if the input image does
            not overlap it. The default value of INDEF does not set a value.

        Returns
        -------
        : tuple
            A tuple with three values: a version string, the number of pixels
            on the input image that do not overlap the output image, and the
            number of complete lines on the input image that do not overlap the
            output input image.

        """

        # Insure that the fillval parameter gets properly interpreted for use with tdriz
        fillval = "INDEF" if util.is_blank(str(fillval)) else str(fillval)
        if insci.dtype > np.float32:
            insci = insci.astype(np.float32)

        # Add input weight image if it was not passed in
        if inwht is None:
            inwht = np.ones_like(insci)

        # Compute what plane of the context image this input would
        # correspond to:
        planeid = int((uniqid - 1) / 32)

        # Check if the context image has this many planes
        if outcon.ndim == 2:
            nplanes = 1
        elif outcon.ndim == 3:
            nplanes = outcon.shape[0]
        else:
            nplanes = 0

        if nplanes <= planeid:
            raise IndexError("Not enough planes in drizzle context image")

        # Alias context image to the requested plane if 3d
        if outcon.ndim == 3:
            outcon = outcon[planeid]

        # Compute the mapping between the input and output pixel coordinates
        # for use in drizzle.cdrizzle.tdriz
        pixmap = resample_utils.calc_gwcs_pixmap(input_wcs, output_wcs, insci.shape)

        log.debug(f"Pixmap shape: {pixmap[:,:,0].shape}")
        log.debug(f"Input Sci shape: {insci.shape}")
        log.debug(f"Output Sci shape: {outsci.shape}")

        log.info(f"Drizzling {insci.shape} --> {outsci.shape}")

        _vers, _nmiss, _nskip = cdrizzle.tdriz(
            insci.astype(np.float32).value,
            inwht,
            pixmap,
            outsci.value,
            outwht.value,
            outcon,
            uniqid=uniqid,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            pixfrac=pixfrac,
            kernel=kernel,
            in_units="cps",
            expscale=1.0,
            wtscale=wtscale,
            fillstr=fillval,
        )


def l2_into_l3_meta(l3_meta, l2_meta):
    """Update the level 3 meta with info from the level 2 meta

    Parameters
    ----------
    l3_meta : dict
        The meta to update. This is updated in-place

    l2_meta : stnode
        The Level 2-like meta to pull from

    Notes
    -----
    The list of meta that is pulled from the Level 2 meta into the Level 3 meta is as follows:
    basic.visit: observation.visit
    basic.segment: observation.segment
    basic.pass: observation.pass
    basic.program: observation.program
    basic.survey: obervation.survey
    basic.optical_element: optical_element
    basic.instrument: instrument.name
    basic.telescope: telescope
    program: program
    """
    l3_meta.basic.visit = l2_meta.observation.visit
    l3_meta.basic.segment = l2_meta.observation.segment
    l3_meta.basic["pass"] = l2_meta.observation["pass"]
    l3_meta.basic.program = l2_meta.observation.program
    l3_meta.basic.survey = l2_meta.observation.survey
    l3_meta.basic.optical_element = l2_meta.instrument.optical_element
    l3_meta.basic.instrument = l2_meta.instrument.name
    l3_meta.coordinates = l2_meta.coordinates
    l3_meta.program = l2_meta.program


def gwcs_into_l3(model, wcs):
    """Update the Level 3 wcsinfo block from a GWCS object

    Parameters
    ----------
    model : `DataModel`
        The model whose meta is to be updated.

    wcs : `GWCS`
        GWCS info to transfer into the `meta.wcsinfo` block

    Notes
    -----
    Some models/parameters in the GWCS object have explicit names, such as
    'crpix1'. However, some do not and hence have to be accessed explicitly
    by indexing. This is fragile and will be a source of issues.
    """
    l3_wcsinfo = model.meta.wcsinfo
    transform = wcs.forward_transform

    l3_wcsinfo.projection = "TAN"
    l3_wcsinfo.pixel_shape = model.shape

    pixel_center = [(v - 1) / 2.0 for v in model.shape[::-1]]
    world_center = wcs(*pixel_center)
    l3_wcsinfo.ra_center = world_center[0]
    l3_wcsinfo.dec_center = world_center[1]
    l3_wcsinfo.pixel_scale_local = compute_scale(wcs, world_center)
    l3_wcsinfo.orientat_local = calc_pa(wcs, *world_center)

    try:
        footprint = utils.create_footprint(wcs, model.shape)
    except Exception as excp:
        log.warning("Could not determine footprint due to %s", excp)
    else:
        l3_wcsinfo.ra_corn1 = footprint[0][0]
        l3_wcsinfo.ra_corn2 = footprint[1][0]
        l3_wcsinfo.ra_corn3 = footprint[2][0]
        l3_wcsinfo.ra_corn4 = footprint[3][0]
        l3_wcsinfo.dec_corn1 = footprint[0][1]
        l3_wcsinfo.dec_corn2 = footprint[1][1]
        l3_wcsinfo.dec_corn3 = footprint[2][1]
        l3_wcsinfo.dec_corn4 = footprint[3][1]
        l3_wcsinfo.s_region = utils.create_s_region(footprint)

    try:
        l3_wcsinfo.x_ref = -transform["crpix1"].offset.value
        l3_wcsinfo.y_ref = -transform["crpix2"].offset.value
    except IndexError:
        log.warning(
            "WCS has no clear reference pixel defined by crpix1/crpix2. Assuming reference pixel is center."
        )
        l3_wcsinfo.x_ref = pixel_center[0]
        l3_wcsinfo.y_ref = pixel_center[1]
    world_ref = wcs(l3_wcsinfo.x_ref, l3_wcsinfo.y_ref)
    l3_wcsinfo.ra_ref = world_ref[0]
    l3_wcsinfo.dec_ref = world_ref[1]
    l3_wcsinfo.pixel_scale = compute_scale(wcs, world_ref)
    l3_wcsinfo.orientat = calc_pa(wcs, *world_ref)

    try:
        l3_wcsinfo.rotation_matrix = transform[
            "pc_rotation_matrix"
        ].matrix.value.tolist()
    except Exception:
        log.warning(
            "WCS has no clear rotation matrix defined by pc_rotation_matrix. Calculating one."
        )
        rotation_matrix = utils.calc_rotation_matrix(l3_wcsinfo.orientat, 0.0)
        l3_wcsinfo.rotation_matrix = utils.list_1d_to_2d(rotation_matrix, 2)


def calc_pa(wcs, ra, dec):
    """Calculate position angle at given ra,dec

    Parameters
    ----------
    wcs : GWCS
        The wcs in consideration.

    ra, dec : float, float
        The ra/dec in degrees.

    Returns
    -------
    position_angle : float
        The position angle in degrees.

    """
    delta_pix = [v for v in wcs.world_to_pixel(ra, dec)]
    delta_pix[1] += 1
    delta_coord = wcs.pixel_to_world(*delta_pix)
    coord = SkyCoord(ra, dec, frame="icrs", unit="deg")

    return coord.position_angle(delta_coord).degree


def populate_mosaic_basic(
    output_model: datamodels.MosaicModel, input_models: [List, ModelContainer]
):
    """
    Populate basic metadata fields in the output mosaic model based on input models.

    Parameters
    ----------
    output_model : MosaicModel
        Object to populate with basic metadata.
    input_models : [List, ModelContainer]
        List of input data models from which to extract the metadata.
        ModelContainer is also supported.

    Returns
    -------
    None
    """

    input_meta = [datamodel.meta for datamodel in input_models]

    # time data
    output_model.meta.basic.time_first_mjd = np.min(
        [x.exposure.start_time.mjd for x in input_meta]
    )
    output_model.meta.basic.time_last_mjd = np.max(
        [x.exposure.end_time.mjd for x in input_meta]
    )
    output_model.meta.basic.time_mean_mjd = np.mean(
        [x.exposure.mid_time.mjd for x in input_meta]
    )

    # observation data
    output_model.meta.basic.visit = (
        input_meta[0].observation.visit
        if len({x.observation.visit for x in input_meta}) == 1
        else -1
    )
    output_model.meta.basic.segment = (
        input_meta[0].observation.segment
        if len({x.observation.segment for x in input_meta}) == 1
        else -1
    )
    output_model.meta.basic["pass"] = (
        input_meta[0].observation["pass"]
        if len({x.observation["pass"] for x in input_meta}) == 1
        else -1
    )
    output_model.meta.basic.program = (
        input_meta[0].observation.program
        if len({x.observation.program for x in input_meta}) == 1
        else "-1"
    )
    output_model.meta.basic.survey = (
        input_meta[0].observation.survey
        if len({x.observation.survey for x in input_meta}) == 1
        else "MULTIPLE"
    )

    # instrument data
    output_model.meta.basic.optical_element = input_meta[0].instrument.optical_element
    output_model.meta.basic.instrument = input_meta[0].instrument.name

    # skycell location
    output_model.meta.basic.location_name = "TBD"

    # association product type
    output_model.meta.basic.product_type = "TBD"


def populate_mosaic_individual(
    output_model: datamodels.MosaicModel, input_models: [List, ModelContainer]
):
    """
    Populate individual meta fields in the output mosaic model based on input models.

    Parameters
    ----------
    output_model : MosaicModel
        Object to populate with basic metadata.
    input_models : [List, ModelContainer]
        List of input data models from which to extract the metadata.
        ModelContainer is also supported.

    Returns
    -------
    None
    """

    input_metas = [datamodel.meta for datamodel in input_models]
    for input_meta in input_metas:
        output_model.append_individual_image_meta(input_meta)



"""Global configuration parameters"""

# Debug mode
#    In conjunction with logging, this mode changes the behavior
#    of functions. Common use case is where associations would be
#    removed, they are no longer removed. This assists in determining
#    why they were an issue.
DEBUG = False



"""
A module that provides functions for matching sky in overlapping images.
"""

import logging
from datetime import datetime

import numpy as np

from .skyimage import SkyGroup, SkyImage

__all__ = ["match"]

__local_debug__ = True

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def match(images, skymethod="global+match", match_down=True, subtract=False):
    """
    A function to compute and/or "equalize" sky background in input images.

    .. note::
       Sky matching ("equalization") is possible only for **overlapping**
       images.

    Parameters
    ----------
    images : list of SkyImage or SkyGroup
        A list of :py:class:`~romancal.skymatch.skyimage.SkyImage` or
        :py:class:`~romancal.skymatch.skyimage.SkyGroup` objects.

    skymethod : {'local', 'global+match', 'global', 'match'}, optional
        Select the algorithm for sky computation:

        * **'local'** : compute sky background values of each input image or
          group of images (members of the same "exposure"). A single sky value
          is computed for each group of images.

          .. note::
            This setting is recommended when regions of overlap between images
            are dominated by "pure" sky (as opposed to extended, diffuse
            sources).

        * **'global'** : compute a common sky value for all input images and
          groups of images. With this setting `local` will compute
          sky values for each input image/group, find the minimum sky value,
          and then it will set (and/or subtract) the sky value of each input image
          to this minimum value. This method *may* be
          useful when the input images have been already matched.

        * **'match'** : compute differences in sky values between images
          and/or groups in (pair-wise) common sky regions. In this case
          the computed sky values will be relative (delta) to the sky computed
          in one of the input images whose sky value will be set to
          (reported to be) 0. This setting will "equalize" sky values between
          the images in large mosaics. However, this method is not recommended
          when used in conjunction with
          `astrodrizzle
          <http://stsdas.stsci.edu/stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_
          because it computes relative sky values while `astrodrizzle` needs
          "absolute" sky values for median image generation and CR rejection.

        * **'global+match'** : first use the **'match'** method to
          equalize sky values between images and then find a minimum
          "global" sky value amongst all input images.

          .. note::
            This is the *recommended* setting for images
            containing diffuse sources (e.g., galaxies, nebulae)
            covering significant parts of the image.

    match_down : bool, optional
        Specifies whether the sky *differences* should be subtracted from
        images with higher sky values (`match_down` = `True`) to match the
        image with the lowest sky or sky differences should be added to the
        images with lower sky values to match the sky of the image with the
        highest sky value (`match_down` = `False`).

        .. note::
          This setting applies *only* when the `skymethod` parameter is
          either `'match'` or `'global+match'`.

    subtract : bool (Default = False)
        Subtract computed sky value from image data.


    Raises
    ------

    TypeError
        The `images` argument must be a Python list of
        :py:class:`~romancal.skymatch.skyimage.SkyImage` and/or
        :py:class:`~romancal.skymatch.skyimage.SkyGroup` objects.


    Notes
    -----

    :py:func:`match` provides new algorithms for sky value computations
    and enhances previously available algorithms used by, e.g.,
    `astrodrizzle
    <http://stsdas.stsci.edu/stsci_python_sphinxdocs_2.13/drizzlepac/astrodrizzle.html>`_.

    Two new methods of sky subtraction have been introduced (compared to the
    standard ``'local'``): ``'global'`` and ``'match'``, as well as a
    combination of the two -- ``'global+match'``.

    - The ``'global'`` method computes the minimum sky value across *all*
      input images and/or groups. That sky value is then considered to be
      the background in all input images.

    - The ``'match'`` algorithm is somewhat similar to the traditional sky
      subtraction method (`skymethod` = `'local'`) in the sense that it
      measures the sky independently in input images (or groups). The major
      differences are that, unlike the traditional method,

        #. ``'match'`` algorithm computes *relative* (delta) sky values with
           regard to the sky in a reference image chosen from the input list
           of images; *and*

        #. Sky statistics are computed only in the part of the image
           that intersects other images.

      This makes the ``'match'`` sky computation algorithm particularly useful
      for "equalizing" sky values in large mosaics in which one may have
      only (at least) pair-wise intersection of images without having
      a common intersection region (on the sky) in all images.

      The `'match'` method works in the following way: for each pair
      of intersecting images, an equation is written that
      requires that average surface brightness in the overlapping part of
      the sky be equal in both images. The final system of equations is then
      solved for unknown background levels.

      .. warning::

        The current algorithm is not capable of detecting cases where some subsets
        of intersecting images (from the input list of images) do not intersect
        at all with other subsets of intersecting images (except for the simple
        case when *single* images do not intersect any other images). In these
        cases the algorithm will find equalizing sky values for each
        intersecting subset of images and/or groups of images.
        However since these subsets of images do not intersect each other,
        sky will be matched only within each subset and the "inter-subset"
        sky mismatch could be significant.

        Users are responsible for detecting such cases and adjusting processing
        accordingly.

    - The ``'global+match'`` algorithm combines the ``'match'`` and
      ``'global'`` methods in order to overcome the limitation of the
      ``'match'`` method described in the note above: it uses the ``'global'``
      algorithm to find a baseline sky value common to all input images
      and the ``'match'`` algorithm to "equalize" sky values in the mosaic.
      Thus, the sky value of the "reference" image will be equal to the
      baseline sky value (instead of 0 in ``'match'`` algorithm alone).

    **Remarks:**
      * :py:func:`match` works directly on *geometrically distorted*
        flat-fielded images thus avoiding the need to perform distortion
        correction on the input images.

        Initially, the footprint of a chip in an image is approximated by a
        2D planar rectangle representing the borders of chip's distorted
        image. After applying distortion model to this rectangle and
        projecting it onto the celestial sphere, it is approximated by
        spherical polygons. Footprints of exposures and mosaics are
        computed as unions of such spherical polygons while overlaps
        of image pairs are found by intersecting these spherical polygons.

    **Limitations and Discussions:**
      Primary reason for introducing "sky match" algorithm was to try to
      equalize the sky in large mosaics in which computation of the
      "absolute" sky is difficult due to the presence of large diffuse
      sources in the image. As discussed above, :py:func:`match`
      accomplishes this by comparing "sky values" in a pair of images in the
      overlap region (that is common to both images). Quite obviously the
      quality of sky "matching" will depend on how well these "sky values"
      can be estimated. We use quotation marks around *sky values* because
      for some image "true" background may not be present at all and the
      measured sky may be the surface brightness of large galaxy, nebula, etc.

      In the discussion below we will refer to parameter names in
      :py:class:`~romancal.skymatch.skystatistics.SkyStats` and these
      parameter names may differ from the parameters of the actual `skystat`
      object passed to initializer of the
      :py:class:`~romancal.skymatch.skyimage.SkyImage`.

      Here is a brief list of possible limitations/factors that can affect
      the outcome of the matching (sky subtraction in general) algorithm:

      * Since sky subtraction is performed on *flat-fielded* but
        *not distortion corrected* images, it is important to keep in mind
        that flat-fielding is performed to obtain uniform surface brightness
        and not flux. This distinction is important for images that have
        not been distortion corrected. As a consequence, it is advisable that
        point-like sources be masked through the user-supplied mask files.
        Values different from zero in user-supplied masks indicate "good" data
        pixels. Alternatively, one can use `upper` parameter to limit the use
        of bright objects in sky computations.

      * Normally, distorted flat-fielded images contain cosmic rays. This
        algorithm does not perform CR cleaning. A possible way of minimizing
        the effect of the cosmic rays on sky computations is to use
        clipping (`nclip` > 0) and/or set `upper` parameter to a value
        larger than most of the sky background (or extended source) but
        lower than the values of most CR pixels.

      * In general, clipping is a good way of eliminating "bad" pixels:
        pixels affected by CR, hot/dead pixels, etc. However, for
        images with complicated backgrounds (extended galaxies, nebulae,
        etc.), affected by CR and noise, clipping process may mask different
        pixels in different images. If variations in the background are
        too strong, clipping may converge to different sky values in
        different images even when factoring in the "true" difference
        in the sky background between the two images.

      * In general images can have different "true" background values
        (we could measure it if images were not affected by large diffuse
        sources). However, arguments such as `lower` and `upper` will
        apply to all images regardless of the intrinsic differences
        in sky levels.

    """
    function_name = match.__name__

    # Time it
    runtime_begin = datetime.now()

    log.info(" ")
    log.info(f"***** {__name__:s}.{function_name:s}() started on {runtime_begin}")
    log.info(" ")

    # check sky method:
    skymethod = skymethod.lower()
    if skymethod not in ["local", "global", "match", "global+match"]:
        raise ValueError(
            "Unsupported 'skymethod'. Valid values are: "
            "'local', 'global', 'match', or 'global+match'"
        )
    do_match = "match" in skymethod
    do_global = "global" in skymethod
    show_old = subtract

    log.info(f"Sky computation method: '{skymethod}'")
    if do_match:
        log.info("Sky matching direction: {:s}".format("DOWN" if match_down else "UP"))

    log.info(
        "Sky subtraction from image data: {:s}".format("ON" if subtract else "OFF")
    )

    # check that input file name is a list of either SkyImage or SkyGroup:
    nimages = 0
    for img in images:
        if isinstance(img, SkyImage):
            nimages += 1
        elif isinstance(img, SkyGroup):
            nimages += len(img)
        else:
            raise TypeError(
                "Each element of the 'images' must be either a "
                "'SkyImage' or a 'SkyGroup'"
            )

    if nimages == 0:
        raise ValueError("Argument 'images' must contain at least one image")

    log.debug(
        "Total number of images to be sky-subtracted and/or matched: {:d}".format(
            nimages
        )
    )

    # Print conversion factors
    log.debug(" ")
    log.debug("----  Image data conversion factors:")

    for img in images:
        img_type = "Image" if isinstance(img, SkyImage) else "Group"

        if img_type == "Group":
            log.debug(f"   *  Group ID={img.id}. Conversion factors:")
            for im in img:
                log.debug(
                    "      - Image ID={}. Conversion factor = {:G}".format(
                        im.id, im.convf
                    )
                )
        else:
            log.debug(f"   *  Image ID={img.id}. Conversion factor = {img.convf:G}")

    # 1. Method: "match" (or "global+match").
    #    Find sky "deltas" that will match sky across all
    #    (intersecting) images.
    if do_match:
        log.info(" ")
        log.info("----  Computing differences in sky values in " "overlapping regions.")

        # find "optimum" sky changes:
        sky_deltas = _find_optimum_sky_deltas(images, apply_sky=not subtract)
        sky_good = np.isfinite(sky_deltas)

        if np.any(sky_good):
            # match sky "Up" or "Down":
            if match_down:
                refsky = np.amin(sky_deltas[sky_good])
            else:
                refsky = np.amax(sky_deltas[sky_good])
            sky_deltas[sky_good] -= refsky

        # convert to Python list and replace numpy.nan with None
        sky_deltas = [skd if np.isfinite(skd) else None for skd in sky_deltas]

        _apply_sky(images, sky_deltas, False, subtract, show_old)
        show_old = True

    # 2. Method: "local". Compute the minimum sky background
    #    value in each sky group/image.
    #    This is an improved (use of masks) replacement
    #    for the classical 'subtract' used by astrodrizzle.
    #
    #    NOTE: incompatible with "match"-containing
    #          'skymethod' modes.
    #
    # 3. Method: "global". Compute the minimum sky background
    #    value *across* *all* sky line members.
    if do_global or not do_match:
        log.info(" ")
        if do_global:
            minsky = None
            log.info(
                '----  Computing "global" sky - smallest sky value '
                "across *all* input images."
            )
        else:
            log.info("----  Sky values computed per image and/or image " "groups.")

        sky_deltas = []
        for img in images:
            sky = img.calc_sky(delta=not subtract)[0]
            sky_deltas.append(sky)
            if do_global and (minsky is None or sky < minsky):
                minsky = sky

        if do_global:
            log.info(" ")
            if minsky is None:
                log.warning('   Unable to compute "global" sky value')
            sky_deltas = len(sky_deltas) * [minsky]
            log.info(
                '   "Global" sky value correction: {} ' "[not converted]".format(minsky)
            )

        if do_match:
            log.info(" ")
            log.info("----  Final (match+global) sky for:")

        _apply_sky(images, sky_deltas, do_global, subtract, show_old)

    # log running time:
    runtime_end = datetime.now()
    log.info(" ")
    log.info(f"***** {__name__:s}.{function_name:s}() ended on {runtime_end}")
    log.info(
        "***** {:s}.{:s}() TOTAL RUN TIME: {}".format(
            __name__, function_name, runtime_end - runtime_begin
        )
    )
    log.info(" ")


def _apply_sky(images, sky_deltas, do_global, do_skysub, show_old):
    for img, sky in zip(images, sky_deltas):
        is_group = not isinstance(img, SkyImage)

        if do_global:
            if sky is None:
                valid = img[0].is_sky_valid if is_group else img.is_sky_valid
                sky = 0.0
            else:
                valid = True

        else:
            valid = sky is not None
            if not valid:
                log.warning(
                    "   *  {:s} ID={}: Unable to compute sky value".format(
                        "Group" if is_group else "Image", img.id
                    )
                )
                sky = 0.0

        if is_group:
            # apply sky change:
            old_img_sky = [im.sky for im in img]
            if do_skysub:
                for im in img:
                    im._image.set_data(im._image.get_data() - sky)
            img.sky += sky
            new_img_sky = [im.sky for im in img]

            # log sky values:
            log.info(
                "   *  Group ID={}. Sky background of "
                "component images:".format(img.id)
            )

            for im, old_sky, new_sky in zip(img, old_img_sky, new_img_sky):
                c = 1.0 / im.convf
                if show_old:
                    log.info(
                        "      - Image ID={}. Sky background: {:G} "
                        "(old={:G}, delta={:G})".format(
                            im.id, c * new_sky, c * old_sky, c * sky
                        )
                    )
                else:
                    log.info(
                        "      - Image ID={}. Sky background: {:G}".format(
                            im.id, c * new_sky
                        )
                    )

                im.is_sky_valid = valid

        else:
            # apply sky change:
            old_sky = img.sky if img.sky is not None else 0
            if do_skysub:
                img._image.set_data(img._image.get_data() - sky)
            if img.sky is None:
                img.sky = 0

            img.sky += sky
            new_sky = img.sky

            # log sky values:
            c = 1.0 / img.convf
            if show_old:
                log.info(
                    "   *  Image ID={}. Sky background: {:G} "
                    "(old={:G}, delta={:G})".format(
                        img.id, c * new_sky, c * old_sky, c * sky
                    )
                )
            else:
                log.info(
                    "   *  Image ID={}. Sky background: {:G}".format(
                        img.id, c * new_sky
                    )
                )

            img.is_sky_valid = valid


def _overlap_matrix(images, apply_sky=True):
    # TODO: to improve performance, the nested loops could be parallelized
    # since _calc_sky() here can be called independently from previous steps.
    ns = len(images)
    data_unit = images[0].image.unit
    A = np.zeros((ns, ns), dtype=float) * data_unit
    W = np.zeros((ns, ns), dtype=float)
    for i in range(ns):
        for j in range(i + 1, ns):
            s1, w1, area1 = images[i].calc_sky(overlap=images[j], delta=apply_sky)

            s2, w2, area2 = images[j].calc_sky(overlap=images[i], delta=apply_sky)
            if area1 == 0.0 or area2 == 0.0 or s1 is None or s2 is None:
                continue

            A[j, i] = s1
            W[j, i] = w1
            A[i, j] = s2
            W[i, j] = w2

    return A, W


def _find_optimum_sky_deltas(images, apply_sky=True):
    ns = len(images)
    A, W = _overlap_matrix(images, apply_sky=apply_sky)
    data_unit = images[0].image.unit

    def is_valid(i, j):
        return W[i, j] > 0 and W[j, i] > 0

    # We need to know how many "non-trivial" (at least for now... - we will
    # compute rank later) equations can be built so that we know the
    # shape of the arrays that need to be created...
    # NOTE: for now use only pairs that *both* have weights > 0 (but a
    # different scenario when only one image has a valid weight can be
    # considered):
    neq = 0
    for i in range(ns):
        for j in range(i + 1, ns):
            if is_valid(i, j):
                neq += 1

    # average weights:
    Wm = 0.5 * (W + W.T)

    # create arrays for coefficients and free terms:
    K = np.zeros((neq, ns), dtype=float)
    F = np.zeros(neq, dtype=float)
    invalid = (ns) * [True]

    # now process intersections between the rest of the images:
    ieq = 0
    for i in range(0, ns):
        for j in range(i + 1, ns):
            if is_valid(i, j):
                K[ieq, i] = Wm[i, j]
                K[ieq, j] = -Wm[i, j]
                F[ieq] = Wm[i, j] * (A[j, i] - A[i, j]).value
                invalid[i] = False
                invalid[j] = False
                ieq += 1

    try:
        rank = np.linalg.matrix_rank(K, 1.0e-12)
    except np.linalg.LinAlgError:
        log.warning("Unable to compute sky: No valid data in common " "image areas")
        deltas = np.full(ns, np.nan, dtype=float)
        return deltas

    if rank < ns - 1:
        log.warning(f"There are more unknown sky values ({ns}) to be solved for")
        log.warning(
            "than there are independent equations available "
            "(matrix rank={}).".format(rank)
        )
        log.warning("Sky matching (delta) values will be computed only for")
        log.warning("a subset (or more independent subsets) of input images.")
    invK = np.linalg.pinv(K, rcond=1.0e-12)

    deltas = np.dot(invK, F)
    deltas[np.asarray(invalid, dtype=bool)] = np.nan
    return deltas * data_unit



import asdf
import astropy.table
import numpy as np
import pytest
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils

from romancal.assign_wcs.assign_wcs_step import load_wcs
from romancal.regtest.regtestdata import compare_asdf


def _add_wcs(tmp_path, model):
    dfn = tmp_path / "wcs_distortion.asdf"
    distortion_model = rdm.DistortionRefModel(maker_utils.mk_distortion())
    distortion_model.save(dfn)
    load_wcs(model, {"distortion": dfn})


@pytest.mark.parametrize("modification", [None, "small", "large"])
def test_compare_asdf(tmp_path, modification):
    # Need to use different directories for the files because the filenames are
    #    now always updated as part of writing a datamodel, see spacetelescope/datamodels#295
    file0 = tmp_path / "test0"
    file1 = tmp_path / "test1"
    file0.mkdir()
    file1.mkdir()

    file0 = file0 / "test.asdf"
    file1 = file1 / "test.asdf"

    l2 = rdm.ImageModel(maker_utils.mk_level2_image(shape=(100, 100)))
    _add_wcs(tmp_path, l2)
    l2.save(file0)
    atol = 0.0001
    if modification == "small":
        l2.data += (atol / 2) * l2.data.unit
    elif modification == "large":
        l2.data += (atol * 2) * l2.data.unit
    l2.save(file1)
    diff = compare_asdf(file0, file1, atol=atol)
    if modification == "large":
        assert not diff.identical, diff.report()
        assert "arrays_differ" in diff.diff
        assert "root['roman']['data']" in diff.diff["arrays_differ"]
    else:
        assert diff.identical, diff.report()


@pytest.mark.parametrize(
    "modification",
    [
        None,
        "dtype",
        "names",
        "small_values",
        "large_values",
        "small_meta",
        "large_meta",
    ],
)
def test_compare_asdf_tables(tmp_path, modification):
    fn0 = tmp_path / "test0.asdf"
    fn1 = tmp_path / "test1.asdf"
    atol = 0.0001
    names = ("a",)
    dtype = np.float32
    values = np.array([1, 2, 3], dtype=dtype)
    t0 = astropy.table.Table([values], names=names)
    # use atol so value is a float
    t0.meta["value"] = atol
    if modification == "names":
        names = ("b",)
    if modification == "small_values":
        values = values + atol / 2
    if modification == "large_values":
        values = values + atol * 2
    if modification == "dtype":
        dtype = np.float64
    t1 = astropy.table.Table([np.array(values, dtype=dtype)], names=names)
    t1.meta["value"] = t0.meta["value"]
    if modification == "small_meta":
        t1.meta["value"] += atol / 2
    if modification == "large_meta":
        t1.meta["value"] += atol * 2
    asdf.AsdfFile({"t": t0}).write_to(fn0)
    asdf.AsdfFile({"t": t1}).write_to(fn1)
    diff = compare_asdf(fn0, fn1, atol=atol)
    if modification in (None, "small_values", "small_meta"):
        assert diff.identical, diff.report()
    else:
        assert not diff.identical, diff.report()
        assert "tables_differ" in diff.diff


def test_model_difference(tmp_path):
    fn0 = tmp_path / "a.asdf"
    fn1 = tmp_path / "b.asdf"
    ma = rdm.DistortionRefModel(maker_utils.mk_distortion())
    mb = rdm.LinearityRefModel(maker_utils.mk_linearity())
    ma.save(fn0)
    mb.save(fn1)
    diff = compare_asdf(fn0, fn1)
    assert not diff.identical
    assert (
        """'type_changes': {"root['roman']": {'new_type': <class 'roman_datamodels.stnode.DistortionRef'>"""
        in diff.report()
    )


@pytest.mark.parametrize("n_diffs", [1, 3, 7])
def test_n_diffs(tmp_path, n_diffs):
    fn0 = tmp_path / "test0.asdf"
    fn1 = tmp_path / "test1.asdf"
    v0 = np.zeros(10, dtype="int32")
    v1 = np.zeros(10, dtype="int32")
    v1[:n_diffs] = 1
    asdf.AsdfFile({"v": v0}).write_to(fn0)
    asdf.AsdfFile({"v": v1}).write_to(fn1)
    diff = compare_asdf(fn0, fn1)
    assert not diff.identical, diff.report()
    assert "arrays_differ" in diff.diff
    assert "root['v']" in diff.diff["arrays_differ"]
    assert n_diffs == diff.diff["arrays_differ"]["root['v']"]["n_diffs"]



"""
Plotting utilities for plotting patches against supplied image

matplotlib dependency is optional.
"""

import spherical_geometry.vector as sgv

import romancal.patch_match.patch_match as pm

try:
    from matplotlib import pyplot as plt
except ImportError:
    print("matplotlib is required for this plotting utility")

plt.ion()


def plot_field(corners, id="", fill=None, color=None):
    plt.fill(corners[0], corners[1], color=fill, edgecolor=color)


def plot_patch(corners, id="", color=None):
    plt.plot(corners[0], corners[1], color=color)
    if id:
        idstr = str(pm.PATCH_TABLE[id]["index"])
        center = (corners[0][:-1].mean(), corners[1][:-1].mean())
        plt.annotate(idstr, center, va="center", ha="center", size=10)


def plot(image_corners, patches_touched_ids, patches_candidate_ids):
    """
    This plots a list of patch footprints against the image footprint.

    Both the touched patchs as well as candidate patches are plotted.


    Parameters
    ----------
    image_corners : Either a squence of 4 (ra, dec) pairs, or
        equivalent 2-d numpy array

    patches_touched_ids: A list of the indices into the patch table of
        patches touched by the image footprint.

    patches_candidate_ids: A list of the indices selected to see if they
        close enough to test if touched by the image footprint.
    """
    plt.clf()
    plt.gca().invert_xaxis()
    plt.plot(0, 0, "*", markersize=10)
    patches_touched = [pm.PATCH_TABLE[index] for index in patches_touched_ids]
    patches_candidate = [pm.PATCH_TABLE[index] for index in patches_candidate_ids]
    tangent_point, patch_tp_id_touched = pm.find_closest_tangent_point(
        patches_touched, image_corners
    )
    ra, dec = sgv.vector_to_lonlat(*tangent_point)
    dummy, patch_tp_id = pm.find_closest_tangent_point(patches_candidate, image_corners)
    vec_image_corners = pm.image_coords_to_vec(image_corners)
    tp_image_corners = pm.veccoords_to_tangent_plane(vec_image_corners, tangent_point)
    plot_field(tp_image_corners, fill="lightgrey", color="black")
    for patch, id in zip(patches_candidate, patches_candidate_ids):
        plot_patch(
            pm.veccoords_to_tangent_plane(
                pm.get_cartesian_corners(patch), tangent_point
            ),
            id=id,
            color="lightgray",
        )
    for patch, id in zip(patches_touched, patches_touched_ids):
        plot_patch(
            pm.veccoords_to_tangent_plane(
                pm.get_cartesian_corners(patch), tangent_point
            ),
            id=id,
            color="blue",
        )
    plt.xlabel("Offset from nearest tangent point in arcsec")
    plt.ylabel("Offset from nearest tangent point in arcsec")
    plt.title(f"RA: {ra} Dec: {dec} of tangent point in degrees")
